




class ValidationError(ValueError):
    '''Inidicates parameters are not valid'''
    pass

class MissingValue(Exception):
    '''Used to indicate that the result should be a missing value.
    This is converted to a status of "missing" in the result.
    '''
    pass

class VulnerabilityError(Exception):
    '''used to indicate that a problem needs to be communicated to the user.
    This is converted to a status of "failure" in the result'''
    pass
    

class Indicator(object):
    '''Specifies a complete transformation from a record in a data set
        (as returned by DataSource.parse) to a raw indicator value
        '''


    def __init__(self):
        raise NotImplemented()

    def getValue(self, asset):
        raw_value = self._getvalue(asset)
        return raw_value

    def _getvalue(self, asset):
        '''must return a value that can be used to get a score
        or raise MissingValue
        or raise Indicator error with a helpful message'''

        raise NotImplemented
        


class Field(Indicator):

    def __init__(self, name, field, missing = None):

        '''
        name - a string describing the indicator
        field - string pointing to the data column
        missing - collection of values to consider as missing
        normalizer - a callable that takes the field value and returns a
        score

        '''

        self.name = name
        self.field = field

        if missing is None:
            missing = []

        self.missing = set(missing)
        self.missing.add(None)


    def _getvalue(self, asset):
        if self.field not in asset.properties:
            msg = 'The asset.properties does not have the key {}.'
            raise VulnerabilityError(msg.format(self.field)) 

        
        raw_value = asset.properties.get(self.field)
        if raw_value in self.missing:
            msg = 'The value for {} is {}, which indicates a missing value.'
            raise MissingValue(msg.format(self.field, raw_value))

        return raw_value
        

class Normalizer(object):

    def __init__(self):
        raise NotImplemented
    

    def getScore(self, raw_value):

        score = self._getscore(raw_value)

        try:
            score = float(score)
        except ValueError:
            msg = 'The score must be convertable to a float. Got {} which is a {}.'
            raise VulnerabilityError(msg.format(score, type(score)))

        if score < 1.0:
            raise VulnerabilityError('the score must be greater than 1')
        if 4.0 < score:
            raise VulnerabilityError('the score must be less than 4')

        return score

    def _getscore(self, raw_value):
        '''should return a score (float between 1 and 4),
        or raise VulnerabilityError with a helpful message'''
        raise NotImplemented
    




class PiecewiseLinearInterpolation(Normalizer):

    def __init__(self, x,y=[1,2,3,4]):
        for _ in (list(x) + list(y)):
            try:
                float(_)
            except ValueError:
                raise ValidationError('{} is not convertable to a float'
                                      ''.format(_))


        self._set_xy(x,y)


    def _set_xy(self, x, y):
        s = [(float(xi), float(yi)) for xi, yi in zip(x,y)]
        s.sort()
        self.x, self.y = zip(*s)


    def _getscore(self, raw_value):

        
        from numpy import interp
        try:
            raw_value = float(raw_value)
        except ValueError:
            raise VulnerabilityError('{} is not convertable to a float'
                                        ''.format(raw_value))
        
        return interp(raw_value, self.x, self.y)
        
        

class DiscreteMapping(Normalizer):

    def __init__(self, mapping):
        self.mapping = mapping



    def _getscore(self, raw_value):
        if raw_value not in self.mapping:
            msg = '{} is not in the discrete mapping'.format(raw_value)
            raise VulnerabilityError(msg)

        return self.mapping[raw_value]


def getWeightedAverage(values, weights):

    s = sum([v * w for v, w in zip(values,weights)])

    return float(s) / sum(weights)

    
class Vulnerability(object):

    components = ['exposure','sensitivity','adaptive capacity']

    def __init__(self, name, weights, indicators):


        self.name = name
        self.weights = weights

        self.indicators = indicators



    def getVulnerability(self, asset):

        try:
            return self._get_vulnerability(asset)
        except MissingValue as e:
            result = VulnerabilityResult()
            result.status = 'missing'
            result.error = e
            return result
        except VulnerabilityError as e:
            result = VulnerabilityResult()
            result.status = 'failure'
            result.error = e
            return result



    def _get_vulnerability(self, asset):
        
        result = VulnerabilityResult()

        c_weights = []
        c_scores = []

        for component, indicators in self.indicators.items():

            i_weights = []
            i_scores = []

            for i_weight, indicator, normalizer in indicators:

                raw_value = indicator.getValue(asset)
                i_score = normalizer.getScore(raw_value)
                

                result.setIndicator(component, raw_value, i_score)

                i_weights.append(i_weight)
                i_scores.append(i_score)

            c_score = getWeightedAverage(i_scores, i_weights)

            result.setComponent(component, c_score)

            c_weights.append(self.weights[component])
            c_scores.append(c_score)

        overall_score = getWeightedAverage(c_scores, c_weights)

        result.setOverall(overall_score)
        result.status = 'success'

        return result
        

        
          
        

class VulnerabilityResult(object):

    def __init__(self):
        self.indicators = dict([(x, []) for x in Vulnerability.components])
        self.components = dict.fromkeys(Vulnerability.components, None)
        self.overall = None

        self.status = None
        self.error = None

    def setIndicator(self, component, raw_value, score):
        self.indicators[component].append((raw_value, score))

    def setComponent(self, component, score):
        self.components[component] = score

    def setOverall(self, score):
        self.overall = score
        
        
        


        
        





        



